from qtpy.QtCore import QThread
from pymodaq.daq_move.utility_classes import DAQ_Move_base
from pymodaq.daq_move.utility_classes import comon_parameters
from pymodaq.daq_utils.daq_utils import ThreadCommand, getLineInfo
from easydict import EasyDict as edict
from ..hardware.piezoconcept.piezoconcept import PiezoConcept, Position, Time

class DAQ_Move_PiezoConcept(DAQ_Move_base):
    """
    Plugin to drive piezoconcpet XY (Z) stages. There is a string nonlinear offset between the set position and the read
    position. It seems to bnot be problem in the sens where a given displacement is maintained. But because the read
    position is not "accurate", I've decided to ignore it and just trust the set position. So the return will be always
    strictly equal to the set position. However, if there is more that 10% difference raise a warning
    """

    _controller_units = 'µm'

    #find available COM ports
    import serial.tools.list_ports
    ports = [str(port)[0:4] for port in list(serial.tools.list_ports.comports())]
    port = 'COM6' if 'COM6' in ports else ports[0] if len(ports) > 0 else ''
    #if ports==[]:
    #    ports.append('')


    is_multiaxes = True
    stage_names = ['X', 'Y', 'Z']
    min_bound = -95  #*µm
    max_bound = 95  #µm
    offset = 100  #µm

    params= [{'title': 'Time interval (ms):', 'name': 'time_interval', 'type': 'int', 'value': 200},
             {'title': 'Controller Info:', 'name': 'controller_id', 'type': 'text', 'value': '', 'readonly': True},
             {'title': 'COM Port:', 'name': 'com_port', 'type': 'list', 'values': ports, 'value': port},
              {'title': 'MultiAxes:', 'name': 'multiaxes', 'type': 'group','visible':is_multiaxes, 'children':[
                        {'title': 'is Multiaxes:', 'name': 'ismultiaxes', 'type': 'bool', 'value': is_multiaxes, 'default': False},
                        {'title': 'Status:', 'name': 'multi_status', 'type': 'list', 'value': 'Master', 'values': ['Master', 'Slave']},
                        {'title': 'Axis:', 'name': 'axis', 'type': 'list',  'values':stage_names},
                        
                        ]}]+comon_parameters

    def __init__(self,parent=None,params_state=None):
        super().__init__(parent, params_state)

        self.controller=None
        self.settings.child(('epsilon')).setValue(1)

    def ini_stage(self, controller=None):
        """
            Initialize the controller and stages (axes) with given parameters.

            =============== =========================================== ===========================================================================================
            **Parameters**   **Type**                                     **Description**

            *controller*     instance of the specific controller object   If defined this hardware will use it and will not initialize its own controller instance
            =============== =========================================== ===========================================================================================

            Returns
            -------
            Easydict
                dictionnary containing keys:
                  * *info* : string displaying various info
                  * *controller*: instance of the controller object in order to control other axes without the need to init the same controller twice
                  * *stage*: instance of the stage (axis or whatever) object
                  * *initialized*: boolean indicating if initialization has been done corretly

            See Also
            --------
            daq_utils.ThreadCommand
        """

        # initialize the stage and its controller status
        # controller is an object that may be passed to other instances of DAQ_Move_Mock in case
        # of one controller controlling multiaxes
        try:
            self.status.update(edict(info="",controller=None,initialized=False))


            #check whether this stage is controlled by a multiaxe controller (to be defined for each plugin)

            # if mutliaxes then init the controller here if Master state otherwise use external controller
            if self.settings.child('multiaxes', 'ismultiaxes').value() and self.settings.child('multiaxes','multi_status').value()=="Slave":
                if controller is None: 
                    raise Exception('no controller has been defined externally while this stage is a slave one')
                else:
                    self.controller = controller
            else: #Master stage
                try:
                    self.close()
                except:
                    pass
                self.controller = PiezoConcept()
                self.controller.init_communication(self.settings.child(('com_port')).value())

            controller_id = self.controller.get_controller_infos()
            self.settings.child(('controller_id')).setValue(controller_id)

            self.settings.child('bounds', 'is_bounds').setValue(True)
            self.settings.child('bounds', 'min_bound').setValue(self.min_bound)
            self.settings.child('bounds', 'max_bound').setValue(self.max_bound)
            self.settings.child('scaling', 'use_scaling').setValue(True)
            self.settings.child('scaling', 'offset').setValue(self.offset)

            self.move_Abs(0)


            self.settings.child(('epsilon')).setValue(2)
            self.status.info = controller_id
            self.status.controller = self.controller
            self.status.initialized = True
            return self.status

        except Exception as e:
            self.emit_status(ThreadCommand('Update_Status', [getLineInfo()+ str(e), 'log']))
            self.status.info = getLineInfo()+ str(e)
            self.status.initialized = False
            return self.status

    def close(self):
        """
            close the current instance of Piezo instrument.
        """
        try:
            self.move_Abs(0)
            QThread.msleep(1000)
        except:
            pass


        self.controller.close_communication()
        self.controller = None


    def check_position(self):
        """
            Check the current position from the hardware.

            Returns
            -------
            float
                The position of the hardware.

            See Also
            --------
            DAQ_Move_base.get_position_with_scaling, daq_utils.ThreadCommand
        """
        position = self.controller.get_position(self.settings.child('multiaxes', 'axis').value())  #in mm
        pos = position.pos/1000  # in um
        pos = self.get_position_with_scaling(pos)
        self.current_position = self.target_position  #should be pos but not precise enough conpared to set position
        self.emit_status(ThreadCommand('check_position', [self.target_position]))
        #print(pos)
        return self.target_position



    def move_Abs(self,position):
        """

        Parameters
        ----------
        position: (float) target position of the given axis in um (or scaled units)

        Returns
        -------

        """
        position = self.check_bound(position)  #limits the position within the specified bounds (-100,100)
        self.target_position = position

        #get positions in controller units
        position = self.set_position_with_scaling(position)
        pos = Position(self.settings.child('multiaxes', 'axis').value(), int(position*1000), unit='n')
        out = self.controller.move_axis('ABS', pos)
        #self.move_is_done = True
        QThread.msleep(50) #to make sure the closed loop converged
        self.poll_moving()


    def move_Rel(self,position):
        """
            Make the hardware relative move of the Piezo instrument from the given position after thread command signal was received in DAQ_Move_main.

            =============== ========= =======================
            **Parameters**  **Type**   **Description**

            *position*       float     The absolute position
            =============== ========= =======================

            See Also
            --------
            DAQ_Move_base.set_position_with_scaling, DAQ_Move_base.poll_moving

        """
        position = self.check_bound(self.current_position+position)-self.current_position
        self.target_position = position+self.current_position

        position = self.set_position_relative_with_scaling(position)

        pos = Position(self.settings.child('multiaxes', 'axis').value(), position*1000, unit='n')  # always use microns for simplicity
        out = self.controller.move_axis('REL', pos)
        QThread.msleep(50)  # to make sure the closed loop converged
        self.poll_moving()

    def move_Home(self):
        """
            Move to the absolute vlue 100 corresponding the default point of the Piezo instrument.

            See Also
            --------
            DAQ_Move_base.move_Abs
        """
        self.move_Abs(100) #put the axis on the middle position so 100µm

    def stop_motion(self):
      """
        Call the specific move_done function (depending on the hardware).

        See Also
        --------
        move_done
      """
      self.move_done()



import importlib
from pathlib import Path
from pymodaq.daq_utils import daq_utils as utils
logger = utils.set_logger('viewer1D_plugins', add_to_console=False)

for path in Path(__file__).parent.iterdir():
    try:
        if '__init__' not in str(path):
            importlib.import_module('.' + path.stem, __package__)
    except Exception as e:
        logger.warning("{:} plugin couldn't be loaded due to some missing packages or errors: {:}".format(path.stem, str(e)))
        pass




# -*- coding: utf-8 -*-
"""
Created on Thu Mar 14 14:54:55 2019

@author: flim-users
"""
import time
import numpy as np
from pymodaq_plugins.hardware.piezoconcept.piezoconcept import PiezoConcept, Position, Time
from pymodaq.daq_utils.daq_utils import set_scan_linear
#%%

stage = PiezoConcept()

stage.init_communication('COM6')
print(stage.get_controller_infos())
stage._write_command('CHAIO 1o2s')
stage._get_read()


#%%
stage.set_time_interval(Time(20,'m'))
stage.get_time_interval()
#%%
stage.get_position('X')
#%%
stage.get_position('Y')
#%%
offset = 100000

start_axis1 = -10000 +offset
start_axis2 = -10000 +offset
stop_axis1 = 10000 +offset
stop_axis2 = 10000 +offset
step_axis1 = 2000
step_axis2 = 2000

scan_params = set_scan_linear(start_axis1,start_axis2,stop_axis1,stop_axis2,step_axis1,step_axis2,back_and_force=True)
positions = scan_params.positions
#%%
stage.move_axis(pos = Position(pos = offset, unit = 'n'))
stage.move_axis(pos = Position(axis = 'Y', pos = offset, unit = 'n'))
stage.set_positions_arbitrary(positions)

#%%
for ind_run in range(1):
    stage.run_arbitrary()
    info= ''
    while 'Completed' not in info:
        time.sleep(0.1)
        info = stage._get_read()
        print(info)
    print(info)
#%%
stage.close_communication()



import pyvisa
import numpy as np

class Position(object):
    units=['n', 'u']
    axes=['X', 'Y']

    def __init__(self,axis='X', pos=100., unit='u'):
        if axis in self.axes:
            self.axis = axis
        else:
            raise Exception('{:s} is not a valid axis'.format(axis))
        self.pos = pos
        if unit in self.units:
            self.unit = unit
        else:
            raise Exception('{:s} is not a valid unit'.format(unit))
        
    def __str__(self):
        return 'Axis {:s} at position {:f}{:s}'.format(self.axis, self.pos, self.unit)    
    
    def __repr__(self):
        return self.__str__()
    
class Time(object):
    units=['u', 'm', 's'] #valid units
    def __init__(self,time=100., unit='u'):
        self.time = time
        if unit in self.units:
            self.unit = unit
        else:
            raise Exception('{:s} is not a valid unit'.format(unit))
            
    def __str__(self):
        return 'Time: {:f}{:s}'.format(self.time, self.unit)    
    
    def __repr__(self):
        return self.__str__()  
    
class PiezoConcept(object):

    def __init__(self):
        super().__init__()
        self._piezo=None
        self._VISA_rm = pyvisa.ResourceManager()
        self.com_ports = self.get_ressources()

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, to):
        self._timeout = to
        self._piezo.timeout = to
        self._piezo.write_timeout = to

    def get_ressources(self):
        infos=self._VISA_rm.list_resources_info()
        com_ports=[infos[key].alias for key in infos.keys()]
        return com_ports
    
    def init_communication(self, com_port):
        if com_port in self.com_ports:
            self._piezo = self._VISA_rm.open_resource(com_port)
            # set attributes
            self._piezo.baud_rate = 115200
            self._piezo.data_bits = 8
            self._piezo.stop_bits = pyvisa.constants.StopBits['one']
            self._piezo.parity = pyvisa.constants.Parity['none']
            self._piezo.flow_control = 0
            self._piezo.read_termination = self._piezo.LF
            self._piezo.write_termination = self._piezo.LF
            self.timeout = 2000
        else:
            raise IOError('{:s} is not a valid port'.format(com_port))
            
    def close_communication(self):
        self._piezo.close()
        self._VISA_rm.close() 
        
    def get_controller_infos(self):
        self._write_command('INFOS')
        return self._get_read()

    def _query(self, command):
        ret = self._piezo.query(command)
        return ret

    def _write_command(self, command):
        self._piezo.write(command)

    
    def _get_read(self):
        self._piezo.timeout = 50
        info = ''
        try:
            while True:
                info += self._piezo.read(encoding='mbcs')+'\n'
        except pyvisa.errors.VisaIOError as e:
            pass
        self._piezo.timeout = self._timeout
        return info
    
    def move_axis(self, move_type='ABS', pos=Position(axis='X', pos=100, unit='n')):
        cmd = '{:s} {:d}{:s}'.format(pos.axis, int(pos.pos), pos.unit)
        if move_type == 'ABS':
            cmd = 'MOVE' + cmd

        elif move_type == 'REL':
            cmd = 'MOVR' + cmd
        else:
            raise Exception('{:s} is not a valid displacement type'.format(move_type))

        ret = self._query(cmd)
        return ret

    def get_position(self, axis='X'):
        """ return the given axis position always in nm
        Parameters
        ----------
        axis: str, default 'X'
            either 'X' or 'Y'
        Returns
        -------
        pos
            an instance of the Position class containing the attributes:
                axis (either ('X' or 'Y'), pos and unit (either 'u' or 'n')
        """    
        pos_str = self._query('GET_{:s}'.format(axis))
        pos_list = pos_str.split(' ')
        unit = pos_list[1][0]
        if unit == 'u':
            pos = int(float(pos_list[0])*1000)
            unit = 'n'
        else:
            pos = int(pos_list[0])

        pos = Position(axis, pos, unit)
        return pos

    def set_time_interval(self, time=Time(50., 'm')):
        if isinstance(time, Time):
            self._query('STIME {:.0f}{:s}'.format(time.time, time.unit))
        else:
            raise Exception('Wrong time argument')
            
    def get_time_interval(self):
        time_str = self._piezo.query('GTIME')
        time_list = time_str.split(' ')
        return Time(float(time_list[0]), time_list[1][0])
    
    def set_positions_simple(self, Xpositions=[], Ypositions=[], Zpositions=[]):
        """ prepare the controller with arbitrary positions in nanometers
        Parameters
        ----------
        Xpositions: (ndarray, dtype=int) X positions in nm
        Ypositions: (ndarray, dtype=int) X positions in nm
        Zpositions: (ndarray, dtype=int) X positions in nm
        """
        Nx = len(Xpositions)
        Ny = len(Ypositions)
        Nz = len(Zpositions)
        
        ret = self._query('ARBWF {:d} {:d} {:d}'.format(Nx, Ny, Nz))
        if ret != 'Ok':
            raise IOError('{:}: Positions not set'.format(ret))
        for xpos in Xpositions:
            ret = self._query('ADDPX {:d}{:s}'.format(int(xpos), 'n'))
            if ret != 'Ok':
                raise IOError('{:}: Added point not set'.format(ret))

        for ypos in Ypositions:
            ret = self._query('ADDPY {:d}{:s}'.format(int(ypos), 'n'))
            if ret != 'Ok':
                raise IOError('{:}: Added point not set'.format(ret))

        for zpos in Zpositions:
            ret = self._query('ADDPZ {:d}{:s}'.format(int(zpos), 'n'))
            if ret != 'Ok':
                raise IOError('{:}: Added point not set'.format(ret))

    def run_simple(self):
        """ run the previously set waveforms
        After writing the command the controller should return the number of steps for each axis, for instance:
        '21.00\n11.00\n2.00\n'
        then :
        'Scan completed\n' once the scan is completed
        check this return using self._get_read(), in order to check when the scan is finished

        """
        self._write_command('RUNWF')


    def set_positions_arbitrary(self, positions):
        """ prepare the controller with arbitrary positions. At least one array should be set
        Parameters
        ----------
        positions: (list) of 2 (or 3 if Z) arbitrary positions
        """

        Npoints = len(positions)
        self._get_read()
        ret = self._query('ARB3D {:d}'.format(Npoints))
        if ret != 'Ok':
            raise IOError('{:}: ARB3D not set'.format(ret))
        
        for pos in positions:
            if len(pos) == 2:
                pos.append(0)
            pos = [int(p) for p in pos]
            ret = self._query('ADD3D {:d}n {:d}n {:d}n'.format(*pos))
            if ret != 'Ok':
                raise IOError('{:}: ADD3D not set'.format(ret))
            
    def run_arbitrary(self):
        self._write_command('RUN3D')
        
    def get_TTL_state(self, port=1):
        """
        Return the configuration state of the given TTL (1 to 4)
        Parameters
        ----------
        port: (int) TTL number (1 to 4)

        Returns
        -------
        list of str : on the form 'status:\nInput rising\nAxis1\n'
        """
        if port > 4 or port < 1:
            raise Exception('Invalid IO port number (1-4)')
        else:
            self._piezo.write('DISIO {:d}'.format(port))
            info = self._get_read()
            info = info.split('\n')
        return info
    
    def set_TTL_state(self, port, axis, IO='disabled', ttl_options=dict(slope='rising', type='start', ind_start=0, ind_stop=0)):
        """ define a given TTL input/output
        Parameters
        ----------
        port: int between 1 and 4
        axis: str either ('X', 'Y' or 'Z')
        IO: str
            either 'disabled', 'input' or 'output'
        ttl_options: dict
            containing the keys:
                slope: str either 'rising' or 'falling' (valid only in 'input' IO mode)
                type: str either 'start', 'end', 'given_step' or 'gate_step' (valid only in 'output' IO mode)
                ind_start: step number to start the TTL (valid for given_step and gate_step mode)
                ind_stop: step number to stop the gate (valid for gate_step mode)
        """
        axes = ['X', 'Y', 'Z']
        ind_axis = axes.index(axis) + 1
        if IO == 'disabled':
            ret = self._write_command('CHAIO {:d}{:s}'.format(port, IO[0]))
        elif IO == 'input':
            ret = self._write_command('CHAIO {:d}{:s}{:d}{:s}'.format(port, IO[0], ind_axis, ttl_options['slope'][0]))
        elif IO == 'output':
            if ttl_options['type'] == 'start':
                ret = self._write_command('CHAIO {:d}{:s}{:d}{:s}'.format(port, 'o', ind_axis, 's'))
            elif ttl_options['type'] == 'end':
                ret = self._write_command('CHAIO {:d}{:s}{:d}{:s}'.format(port, 'o', ind_axis, 'e'))
            elif ttl_options['type'] == 'given_step':
                ret = self._write_command('CHAIO {:d}{:s}{:d}{:s}{:d}'.format(port, 'o', ind_axis, 'n', ttl_options['ind_start']))
            elif ttl_options['type'] == 'gate_step':
                ret = self._write_command('CHAIO {:d}{:s}{:d}{:s}{:d}-{:d}'.format(port, 'o', ind_axis, 'g', ttl_options['ind_start'],
                                                             ttl_options['ind_stop']))

        else:
            raise Exception('Not valid IO type for TTL')

        info = self._get_read()
        info = info.split('\n')

        if info[-2] != 'Ok':
            raise IOError('{:}: set_TTL_state wrong return'.format(ret))

class PiezoConceptPI(PiezoConcept):
    def __init__(self):
        super().__init__()

    def get_controller_infos(self):
        self._write_command('*IDN?')
        return self._get_read()

    def get_position(self, axis='X'):
        """ return the given axis position always in nm
        Parameters
        ----------
        axis: str, default 'X'
            either 'X' or 'Y'
        Returns
        -------
        pos
            an instance of the Position class containing the attributes:
                axis (either ('X' or 'Y'), pos and unit (either 'u' or 'n')
        """
        ind = ['X', 'Y'].index(axis) + 1
        pos_str = self._query(f'POS? {ind}')
        pos = Position(axis, float(pos_str), 'u')
        return pos

    def move_axis(self, move_type='ABS', pos=Position(axis='X', pos=100., unit='u')):
        if pos.unit == 'n':
            pos.pos = pos.pos / 1000
            pos.unit = 'u'
        ind = ['X', 'Y'].index(pos.axis) + 1
        cmd = f'{ind} {pos.pos}'
        if move_type == 'ABS':
            cmd = 'MOV ' + cmd
        else:
            raise Exception('{:s} is not a valid displacement type'.format(move_type))

        ret = self._write_command(cmd)
        return ret


from pymodaq_plugins.hardware.piezoconcept.piezoconcept import PiezoConcept, Position, Time
import numpy as np

if __name__ == '__main__':

    controller = PiezoConcept()
    controller.init_communication('COM6')
    print(controller.get_controller_infos())
    controller.get_time_interval()
    xaxis = np.linspace(100000, 120000, 21)
    yaxis = np.linspace(100000, 110000, 11)
    zaxis = np.linspace(0, 0, 11)

    #controller.set_positions_simple(xaxis, yaxis, [])
    #controller.run_simple()
    #controller._get_read()
    xaxis = np.linspace(100000, 120000, 21)
    yaxis = np.linspace(100000, 110000, 21)
    #controller.set_positions_arbitrary([xaxis, yaxis])
    #controller.run_arbitrary()
    #controller._get_read()
    controller.get_TTL_state(1)
    controller._get_read()
    pass